/* Copyright (c) 2014-2017 Louisiana State University
 *               2017-2025 Cardiff University
 *
 * This file is part of GWpy.
 *
 * GWpy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GWpy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * PyData Sphinx theme overrides
 */

/* Colours */
html[data-theme="light"] {
		--pst-color-secondary: #b53c00;
}
html[data-theme="dark"] {
		--pst-color-secondary: #b53c00;
}

/* Hide left sidebar title */
nav.bd-links p.bd-links__title {
		display: none;
}
nav.bd-links p.caption {
		font-size: var(--pst-sidebar-header-font-size);
}

/* Dedent sidebar links */
nav.bd-links ul.bd-sidenav a {
		padding-left: 0;
}
nav.bd-links .current > a {
		box-shadow: none;
}

/* In-Page TocTree styling */
.toctree-wrapper-li[class^=toctree-l] > a {
		font-size: 1em;
}

/*
 * Link styling
 */

a.image-reference {
		text-decoration: none;
}

/*
 * Code signature styling
 */

.sig .sig-param {
		font-family: var(--pst-font-family-monospace);
		font-style: normal;
}

dl.field-list span.classifier {
		font-style: normal;
		font-weight: 400;
}

/*
 * Sphinx-gallery
 */

div.sphx-glr-download-link-note {
		display: none !important;
}
